/*
 * Copyright (c) 2003, Raymond Bosman
 * Copyright (c) 2003, Frederik Holljen
 * All Rights Reserved.
 *
 * See COPYING for licensing.
 */


header "post_include_hpp"
{
	#include "../antlr/CkASTNodeBase.hpp"

	// Since namspace not implemented, we use defines.
	#define Undefined CkDataType::Undefined
	#define PrimVoid CkDataType::PrimVoid
	#define PrimBool CkDataType::PrimBool
	#define PrimChar CkDataType::PrimChar
	#define PrimInt CkDataType::PrimInt
	#define Custom CkDataType::Custom

	#include "../tools/symboltable.hpp"
}

header "post_include_cpp"
{
	#include <iostream>
    #include <sstream>
    #include "../tools/datatype.hpp"
    #include "../tools/errorhandler.hpp"
    #include "../tools/exception.hpp"
	using namespace std;
}

// Global options
options
{
	mangleLiteralPrefix = "TK_"; // instead of TOKEN
	language="Cpp";
}

/**
 * \brief Checker class
 * This class is a treewalker that checks the parsed tree. Some tree modifications
 * are made in this class. Extra node information is needed by the codegenerator.
 */
class CkChecker extends TreeParser;
options
{
	importVocab=Ck;
    buildAST = true;     // "transform" mode
    ASTLabelType = "RefCkASTNodeBase";
}

//! Define stuff we need in the class header
{
 private:
    SymbolTable<CheckerIdEntry> Symbol_table;
    SymbolTable<FunctionIdEntry> Function_table;
	CkDataType::Types functionReturn;
	int programReturn;


	/* \brief Return a CkDataType::Types from the ANTLR generated types. 
	 * \param ANTLR generated type.
	 */
	CkDataType::Types getResultType (int ANTLRType)
	{
		switch( ANTLRType )
        {
		     case PRIM_BOOLEAN: return PrimBool;  break;
    	     case PRIM_CHAR:  	return PrimChar;  break;
        	 case PRIM_INT:  	return PrimInt;   break;
	         case PRIM_VOID: 	return PrimVoid;  break;
			 default:			return Undefined; break;
		}
	};
 
}

/**
 * \brief Top node in the AST tree. Starts treewalking by visiting all the function declarations.
 */
program
   : #(PROGRAM (parseFunction)*)
   ;

/**
 * \brief Called from program, checks the function declaration.
 *
 * This function has the following tasks:
 * - Open new scope for the parameters within the function. 
 * - Check the return type of a block with the given (primitive) type of the function.
 * - Add the function identifier and it's parametesr to the function_table. 
 */
parseFunction
	: #(FUNCTION_DECL 
	{	
		// Strange results appear if we do the openscope at a later time.
		Symbol_table.openScope();
	} 
	prim:primitive
	{
		// Set the return type of the function. 
		// This holds for every function and including main.
		functionReturn = getResultType(prim->getType());
	}
	( id:IDENT 
	{
		// Create the function ID.
        FunctionIdEntry *func = new FunctionIdEntry( prim->getLine() );

		// And set the resultType.
		func->ResultType = getResultType(prim->getType());  
	}
	
	(par:parameter
	{
		// Add all parameters.
		func->paramAdd(par->getResultType());
	}
	)* 
	{
		// After the parameters
    	Function_table.enter( id->getText(), func);
	}
	b:block
	{
	 	// Check the return (or last) expression of a block. As long as it is not a
		// PrimVoid, check it's result. 
		if(functionReturn != PrimVoid && (b->getResultType() != functionReturn ))
		{
			stringstream error;
			error << "Invalid end of block. Need an expression of type "
				<< CkDataType::dataTypeToString(functionReturn);
            ErrorHandler::error( b->getLine(), error.str() );
		}
	}
	
	| MAIN block)
	{
		// End of the function, we can close the scope.
        Symbol_table.closeScope();
	}
	)
	;

/**
 * \brief Check a function parameter declaration.
 * 
 * This function checks if the parameter is already declared at this scope (duplicated parameters).
 * If not the function adds the parameter to the symbol table. 
 */
parameter
	: #(pd:PARAM_DEF (cnst:CONST)? prim:primitive {CkDataType type;} type=id:varDeclaration)
	{
		// Add the parameters to the current scope. 
        CkDataType::Types varType = getResultType(prim->getType()); 

        try
        {
        	type.setReturnType( varType );

			// Add parameter to the symbol table. 
        	Symbol_table.enter( id->getText(),
                    new CheckerIdEntry( prim->getLine(), type ,cnst ? true: false ) );

			pd->setResultType( type ); // Set the param_def too.
			
        } catch( SymbolTableException & )
        {
        	stringstream error;
            error << "Identifier " << id->getText()
                  << " already declared in this scope at line: "
                  << Symbol_table.retrieve( id->getText() )->LineNr;
            ErrorHandler::error( id->getLine(),  error.str() );
        }
	}
	;

/**
 * \brief Checks a block ( code in { } ).
 * 
 * This function has the following tasks:
 * - Open new scope for a new block.
 * - Check the last expression, and see if it returns the correct type.
 * - If the type is correct or cannot checked at this time, type information is added.
 *
*/
block
    : #(b:BLOCK
    {
		// Do not open a new scope if we are checking a 'special case'. 
		// (if and while-blocks)
        if( !b->getIgnoreScope() ) Symbol_table.openScope();

    } (declaration | ex:expression | retExpr:returnExpr)*)
	{
		// Last expression of the block
    	if( ex && ex->getNextSibling() == 0 )
        {
        	if( ex->getResultType() == Undefined )
			{
                    ErrorHandler::error( ex->getLine(), "Return expression has type: " +
                                         ex->getResultType().toString() );
			}
										 
        	b->setResultType( ex->getResultType() );
				
			// Set the line number of the block, to this statement. The next check
			// will be made after finishing this block.
			b->setLine(ex->getLine());
        }
        else if( retExpr && retExpr->getNextSibling() == 0 )// retExpr is last
        {
        	if( retExpr->getResultType() == PrimVoid || retExpr->getResultType() == Undefined )
                         ErrorHandler::error( retExpr->getLine(), "Return expression has type: " +
                                              retExpr->getResultType().toString() );
            b->setResultType( retExpr->getResultType() );
        }

        try
        {
			// And close the scope again
        	if( !b->getIgnoreScope() ) Symbol_table.closeScope();
        }
		catch( SymbolTableException & )
        {
            throw Exception( __FILE__, __LINE__, "Compiler oops: Unmatched scopes" );
   		}
	}
	;

/*
 * \brief Check if the expression after the return statement is of a correct type. 
 *
 * \todo return statement is not fully implemented. Functions don't have to end with a
 * return. (Although it's hard to check if every brench actually returns.)
 */ 
returnExpr
	: #(ret:RETURN ex:expression)
	{
		//cout << "Warning: Return statements are not fully implemented" << endl;
		
		if (functionReturn != PrimVoid && ex->getResultType() != functionReturn)
		{
			stringstream error;
			error << "Invalid return type: got " << ex->getResultType().toString() << 
				" and expected " << CkDataType::dataTypeToString(functionReturn);
            ErrorHandler::error( ex->getLine(), error.str() );
		}

        ret->setResultType( ex->getResultType() );
	}
	;



/**
 * \brief Check the common parts of a variable declaration.
 * 
 * This function has the following tasks:
 * - Check if the symbol is already declared.
 * - Add the defined variable to the symbol table.
 *
 * A variable declaration can be followed by a assignment.
 */
declaration
	: #(DECLARATION (cnst:CONST)? prim:primitive {CkDataType type;}
                                     type=id:varDeclaration (expr:expression)? (ai:arrayInit)? )
	{
        // register var
        CkDataType::Types varType = getResultType(prim->getType());
		
		// Check the varType.
		if (varType == PrimVoid || varType == Undefined)
		{
	       ErrorHandler::error( prim->getLine(),
   		        "Can't declare variable " + id->getText() +
                " of type void or undefined" );
		}

        try
        {
	        type.setReturnType( varType );
            Symbol_table.enter( id->getText(),
                    new CheckerIdEntry( prim->getLine(), type ,cnst ? true: false ) );

            id->setResultType( varType );
        }
		catch( SymbolTableException & )
        {
            stringstream error;
            error << "Identifier " << id->getText()
                  << " already declared in this scope at line: "
                  << Symbol_table.retrieve( id->getText() )->LineNr;
            ErrorHandler::error( id->getLine(),  error.str() );
        }

        if( expr &&  expr->getResultType() != type ) // init of single
        {
            ErrorHandler::error( prim->getLine(), "No match for " + type.toString()
                                 + " = " + expr->getResultType().toString() + " expression" );
        }
        else if( ai && ai->getResultType() != type ) // init of array
        {
            ErrorHandler::error( prim->getLine(), "No match for "
                                 + type.toString() + " = " + ai->getResultType().toString()
                                 + " initialisation" );
        }
	}
	;

/** 
 * \brief Check the variable declaration. Serious array checking implemented in 
 * this function. 
 * 
 * A variable can be declared in a block OR as a parameter.
 * 
 * Block: The parser expects: IDENT and possibly followed by brackets. Between the
 *        brackets an integer is expected. The integer gives the size of the array. 
 *        The array has a minimum of 1 and a maximum of 2 dimensions. A 0 dimensional
 *        array is not possible, otherwise it was just an ordinary IDENT. 
 *
 * Param: When the ARRAY is defined as parameter, we don't have information about the
 *        size. By a singe dimensional array, on sizes are given. With a multi 
 *        dimensional array is only the '[' token given. (Some people call this a hack
 *        (including me)).
 *      
 */                              
varDeclaration
returns [CkDataType type]
    : IDENT
    | #(ar:ARRAY id:IDENT 
	{
	    int dim1, dim2;

        dim1 = -1; // Means 1 dimension, with no certain boundary (function parameter). 
		dim2 = 0;
		// Both can be overwritten.
	}
	
	(LBRACKET
	{
		// We got only a Left Bracket when it is a multi dimensional, non boundary array.
		dim2 = -1;
	}
	)?
	
	(d1:INT (d2:INT)? )? )
    {
        ar->setText( id->getText() ); // for easy symbol table add in declaration
        ar->setLine( id->getLine() ); // make error msgs correct

        if( d1 )
        {
        	stringstream ss( d1->getText() );
	        ss >> dim1;

        	if( dim1 < 1 )
    	    {
            	ErrorHandler::error( id->getLine(), "Cant declare array "
                                 + id->getText() + " with 0 dimension" );
	        }
		}

        if( d2 )
        {
            stringstream ss( d2->getText() );
            ss >> dim2;
            if( dim2 < 1 )
            {
                ErrorHandler::error( id->getLine(), "Cant declare array "
                                     + id->getText() + " with 0 dimension" );
            }
        }
        type.setDimensions( dim1, dim2 );
    }
    ;

/**
 * \brief Checks an array initialisation for either single or double arrays.
 * 
 * This function checks the types and the dimensions of the initialisation part. 
 * 
 * \internal We cheat and use operand. We can do this since we know there will 
 * never be a ident (parser checks this). A cleaner solution is ofcourse to split 
 * operand into two parts.
 */
arrayInit
    : #(si:SINGLE_ARRAY_INIT 
	{ 
		int count = 0; CkDataType type; 
	} 
	(op:operand
    {
    	count++;
        if( type.getReturnType() == Undefined ) // first
        {
        	type =  op->getResultType();
        }
        else // check that params are of same type
        {
        	if( op->getResultType() != type )
            {
            	ErrorHandler::error( op->getLine(), "Different types "
                	+ type.toString() + " and "
                    + op->getResultType().toString()
                    + " in array initialisation" );
            }
        }
    }
    )+ )
    {
    	si->setLine( op->getLine() ); // make error msgs correct
    	type.setDimensions( count );
        si->setResultType( type );
    }
    | #(di:DOUBLE_ARRAY_INIT 
	{ int count = 0; CkDataType type; } 
	(single:arrayInit
    {
    	count++;
        if( type.getReturnType() == Undefined ) // first
        {
        	type = single->getResultType();
        }
        else // check that params are of same type
        {
        	if( single->getResultType() != type )
            {
            	ErrorHandler::error( single->getLine(), "Different types "
                	+ type.toString() + " and "
                    + single->getResultType().toString()
                    + " in array initialisation" );
            }
        }
    }
    )+ )
    {
    	type.expandDimension( count );
        di->setResultType( type );
    }
    ;

/**
 * \brief The main checking method for expressions.
 * 
 * Checking is done if two (sub)results (left-hand and right-hand) not conflict 
 * within the expression.
 */
expression
	/* highest priority */
	: assign

/*!
 * \brief || (or) operation (priority 6)
 * 
 * boolean || boolean ==> boolean
 */ 
    | #(orOp:OR lh:expression rh:expression) /* ** OR ** */
    {
        if( lh->getResultType() != PrimBool || rh->getResultType() != PrimBool
            || lh->getResultType().getNumDimensions() != 0 || rh->getResultType().getNumDimensions() != 0 )
        {
            ErrorHandler::error( orOp->getLine(), "No match for "
                                 + lh->getResultType().toString()
                                 + " || " + rh->getResultType().toString()
                                 + " operator" );
        }
        orOp->setResultType( PrimBool );
    }

/*!
 * \brief && (and) operation (priority 5)
 * 
 * boolean && boolean ==> boolean
 */ 
	| #(andOp:AND andlh:expression andrh:expression) /* ** AND ** */
    {
        if( andlh->getResultType() != PrimBool || andrh->getResultType() != PrimBool
            || andlh->getResultType().getNumDimensions() != 0
            || andrh->getResultType().getNumDimensions() != 0 )
        {
            ErrorHandler::error( andOp->getLine(), "No match for "
                                 + andlh->getResultType().toString()
                                 + " && " + andrh->getResultType().toString()
                                 + " operator" );
        }
        andOp->setResultType( PrimBool );
    }

/*!
 * \brief < (less) operation (priority 4)
 * 
 * int < int ==> boolean
 */ 
    | #(ltOp:LTHEN ltlh:expression ltrh:expression) /* ** LT ** */
    {
        if( ltlh->getResultType() != PrimInt || ltrh->getResultType() != PrimInt
            || ltlh->getResultType().getNumDimensions() != 0
            || ltrh->getResultType().getNumDimensions() != 0 )
        {
            ErrorHandler::error( ltOp->getLine(), "No match for "
                                 + ltlh->getResultType().toString()
                                 + " < " + ltrh->getResultType().toString()
                                 + " operator" );
        }
        ltOp->setResultType( PrimBool );
    }

/*!
 * \brief <= (less or equal) operation (priority 4)
 * 
 * int < int ==> boolean
 */ 
 	| #(leOp:LEQUAL lelh:expression lerh:expression) /* ** LEQUAL ** */
    {
        if( lelh->getResultType() != PrimInt || lerh->getResultType() != PrimInt
            || lelh->getResultType().getNumDimensions() != 0
            || lerh->getResultType().getNumDimensions() != 0 )
        {
            ErrorHandler::error( leOp->getLine(), "No match for "
                                 + lelh->getResultType().toString()
                                 + " <= " + lerh->getResultType().toString()
                                 + " operator" );
        }
        leOp->setResultType( PrimBool );
    }

/*!
 * \brief > (greater) operation (priority 4)
 * 
 * int > int ==> boolean
 */ 
    | #(gtOp:GTHEN gtlh:expression gtrh:expression) /* ** GTHEN ** */
    {
        if( gtlh->getResultType() != PrimInt || gtrh->getResultType() != PrimInt
            || gtlh->getResultType().getNumDimensions() != 0
            || gtrh->getResultType().getNumDimensions() != 0 )
        {
            ErrorHandler::error( gtOp->getLine(), "No match for "
                                 + gtlh->getResultType().toString()
                                 + " > " + gtrh->getResultType().toString()
                                 + " operator" );
        }
        gtOp->setResultType( PrimBool );
    }

/*!
 * \brief >= (greater or equal) operation (priority 4)
 * 
 * int >= int ==> boolean
 */ 
    | #(geOp:GEQUAL gelh:expression gerh:expression) /* ** GEQUAL ** */
    {
        if( gelh->getResultType() != PrimInt || gerh->getResultType() != PrimInt
            || gelh->getResultType().getNumDimensions() != 0
            || gerh->getResultType().getNumDimensions() != 0 )
        {
            ErrorHandler::error( geOp->getLine(), "No match for "
                                 + gelh->getResultType().toString()
                                 + " >= " + gerh->getResultType().toString()
                                 + " operator" );
        }
        geOp->setResultType( PrimBool );
    }

/*!
 * \brief == (equals) operation (priority 4)
 * 
 * int == int ==> boolean
 * char == char ==> boolean
 * bool == bool ==> boolean
 */ 
    | #(eqOp:EQUAL eqlh:expression eqrh:expression) /* ** EQUAL ** */
    {
        if( eqlh->getResultType() != eqrh->getResultType() ||  // not equal
            ( eqlh->getResultType() != PrimBool &&             // not a known type
              eqlh->getResultType() != PrimChar &&
              eqlh->getResultType() != PrimInt ) )
        {
            ErrorHandler::error( eqOp->getLine(), "No match for "
                                 + eqlh->getResultType().toString()
                                 + " == " + eqrh->getResultType().toString()
                                 + " operator" );
        }
        eqOp->setResultType( PrimBool );
    }

/*!
 * \brief != (not) operation (priority 4)
 * 
 * int != int ==> boolean
 * char != char ==> boolean
 * bool != bool ==> boolean
 */ 
    | #(neOp:NEQUAL nelh:expression nerh:expression) /* ** NEQUAL ** */
    {
        if( nelh->getResultType() != nerh->getResultType() ||  // not equal
            ( nelh->getResultType() != PrimBool &&             // not a known type
              nelh->getResultType() != PrimChar &&
              nelh->getResultType() != PrimInt ) )
        {
            ErrorHandler::error( neOp->getLine(), "No match for "
                                 + nelh->getResultType().toString()
                                 + " != " + nerh->getResultType().toString()
                                 + " operator" );
        }
        neOp->setResultType( PrimBool );
    }

/*!
 * \brief + (plus) operation (priority 3 or priority 0)
 * 
 * normal plus lh + rh has priority 3,
 * unary plus +lh has priority 0.
 *
 * int + int ==> int
 * + int ==> int
 */ 
    | #(plusOp:PLUS pluslh:expression /* UNARY or expression PLUS */
	{
		if( pluslh->getResultType() != PrimInt || 
			pluslh->getResultType().getNumDimensions() != 0 )
		{
			ErrorHandler::error( plusOp->getLine(), "No match for "
                                 + pluslh->getResultType().toString()
                                 + " operator" );
		}
	    plusOp->setResultType( PrimInt );
	}

	(plusrh:expression /* expression PLUS */
	{
		if( plusrh->getResultType() != PrimInt || 
			plusrh->getResultType().getNumDimensions() != 0 )
		{
			ErrorHandler::error( plusOp->getLine(), "No match for "
                                 + pluslh->getResultType().toString()
                                 + " + " + plusrh->getResultType().toString()
                                 + " operator" );
		}
	    plusOp->setResultType( PrimInt );
	}
	)?)

/*!
 * \brief - (minus) operation (priority 3 or priority 0)
 * 
 * normal minus lh - rh has priority 3,
 * unary minus -lh has priority 0.
 *
 * int - int ==> int
 * - int ==> int
 */ 
    | #(minusOp:MINUS minuslh:expression  /* UNARY or expression MINUS */
		{
	        if( minuslh->getResultType() != PrimInt || minuslh->getResultType().getNumDimensions() != 0 )
	    ErrorHandler::error( minusOp->getLine(), "No match for "
                                 + pluslh->getResultType().toString()
                                 + " operator" );
	    minusOp->setResultType( PrimInt );
	}
	(minusrh:expression /* expression MINUS */
		{
	        if( minusrh->getResultType() != PrimInt || minusrh->getResultType().getNumDimensions() != 0 )
	    ErrorHandler::error( minusOp->getLine(), "No match for "
                                 + pluslh->getResultType().toString()
                                 + " - " + plusrh->getResultType().toString()
                                 + " operator" );
	    minusOp->setResultType( PrimInt );
		}
	)?)

/*!
 * \brief * (multiply) operation (priority 2)
 * 
 * int * int ==> int
 */ 
	| #(multOp:MULTIPLY multlh:expression multrh:expression)
    {
        if( multlh->getResultType() != PrimInt || multrh->getResultType() != PrimInt
            || multlh->getResultType().getNumDimensions() != 0
            || multrh->getResultType().getNumDimensions() != 0 )
        {
            ErrorHandler::error( multOp->getLine(), "No match for "
                                 +  multlh->getResultType().toString()
                                 + " * " + multrh->getResultType().toString()
                                 + " operator" );
        }
        multOp->setResultType( PrimInt );
    }

/*!
 * \brief / (divide) operation (priority 2)
 * 
 * int / int ==> int
 */ 
	| #(divOp:DIVIDE divlh:expression divrh:expression)
    {
        if( divlh->getResultType() != PrimInt || divrh->getResultType() != PrimInt
            || divlh->getResultType().getNumDimensions() != 0
            || divrh->getResultType().getNumDimensions() != 0 )
        {
            ErrorHandler::error( divOp->getLine(), "No match for "
                                 + divlh->getResultType().toString()
                                 + " / " + divrh->getResultType().toString()
                                 + " operator" );
        }
        divOp->setResultType( PrimInt );
    }

/*!
 * \brief % (modulo) operation (priority 2)
 * 
 * int / int ==> int
 */ 
	| #(modOp:MODULO modlh:expression modrh:expression)
    {
        if( modlh->getResultType() != PrimInt || modrh->getResultType() != PrimInt
            || modlh->getResultType().getNumDimensions() != 0
            || modrh->getResultType().getNumDimensions() != 0 )
            ErrorHandler::error( modOp->getLine(), "No match for "
                                 + modlh->getResultType().toString()
                                 + " / " + modrh->getResultType().toString()
                                 + " operator" );
        modOp->setResultType( PrimInt );
    }

/*!
 * \brief ! (not) operation (priority 1)
 * 
 * ! bool ==> bool
 */ 
    | #(notOp:NOT notExpr:expression) /* ** NOT ** */
    {
        if( notExpr->getResultType() != PrimBool ||
            notExpr->getResultType().getNumDimensions() != 0 )
        {
            ErrorHandler::error( notOp->getLine(), "No match for ! "
                                 + notExpr->getResultType().toString()
                                 + " operator" );

        }
        notOp->setResultType( PrimBool );
    }

	/* Priority 0 */
	| operand
	| block
	| paren

	/* And expression functions */
	| getStatement
	| putStatement
	| ifStatement
	| whileStatement
	;

/**
 * \brief Set the resultType of the expression between parens.
 */
paren
	: #(par:PAREN parExpr:expression)
	{
		par->setResultType (parExpr->getResultType());
	}
	;

/**
 * \brief Set the types of the operands. A higher hierarchie checks if  
 * they conflict.
 */
operand
	: variable
	| i:INT
      {
          i->setResultType( PrimInt );
      }
	| c:CHAR
      {
          c->setResultType( PrimChar );
      }
	| bt: TRUE
	  {
	  		bt->setResultType ( PrimBool );
	  }
	| bf: FALSE
	  {
	  		bf->setResultType ( PrimBool );
	  }
	;

/**
 * \brief Check the type of a variable, array variable and function call.
 * 
 * A array variable needs also a dimension check.
 * The function call, also checks the existens of the function, the 
 * number of parameters, the types of those parameters. 
 */
variable
	: id:IDENT
		{
			// Check for declaration.
            if( Symbol_table.retrieve( id->getText() ) == 0 )
            {
                ErrorHandler::error( id->getLine(), "Undefined variable " + id->getText() );
            }
            else
            {
                // clean var, same type as original definition
                id->setResultType( Symbol_table.retrieve( id->getText() )->Type );
            }
		}
    | #(a:ARRAY arrayId:IDENT expr1:expression (expr2:expression)? )
    {
        a->setText( arrayId->getText() ); // make array identifiable
        a->setLine( arrayId->getLine() ); // make error msgs correct
        if( expr1->getResultType() != PrimInt )
        {
            ErrorHandler::error( expr1->getLine(),
                                 "No match for " + arrayId->getText() + "[ " +
                                 expr1->getResultType().toString() + " ].." );
        }
        if( expr2 && expr2->getResultType() != PrimInt )
        {
            ErrorHandler::error( expr2->getLine(),
                                 "No match for " + arrayId->getText() + "[ " +
                                 expr1->getResultType().toString() + " ][ "
                                 + expr2->getResultType().toString() + " ]");
        }
        // Check for declaration.
        CheckerIdEntry *entry = Symbol_table.retrieve( arrayId->getText() );
        if( entry == 0 )
        {
            ErrorHandler::error( arrayId->getLine(), "Undefined variable " + arrayId->getText() );
        }
        else
        {
            // identifer usage has array operators, this means reduce of original type
            // throws if we try to perform an illegal reduction
            try
            {
                a->setResultType( entry->Type.reduceDimensions( expr2 ? 2 : 1 ) );
            }catch( const Exception &e )
            {
                ErrorHandler::error( arrayId->getLine(), e.getError() );
            }
        }
    }

    | #(func:FUNCTION funcID:IDENT 
	{
			FunctionIdEntry *entry = Function_table.retrieve(funcID->getText());

			if( entry == 0 )
	        {
    	        ErrorHandler::error( funcID->getLine(), "Undefined function " + funcID->getText() );
			}
			else
			{
				try
				{
					// Set the return type.
					func->setResultType( entry->ResultType );
				}catch( const Exception &e )
				{
					ErrorHandler::error( funcID->getLine(), e.getError() );
				}

				// Parameter iterator reset.
				entry->paramReset();
			}
 	}
	{int paramCount = 0;
	} (paramType:expression
	{
		if(entry != 0)
		{
		// Check the parameter types.
		paramCount++;

		if (paramCount > entry->paramSize())
		{
		  	stringstream error;
			error << "Function caller has too many parameters ";
			ErrorHandler::error( paramType->getLine(), error.str());
		}
       
			// Get the declared parameter.
			CkDataType declared = entry->paramGet();

			// Compare it with this Type.
			
			// When we have a identifier as input.
			if (paramType->getType() == IDENT) 
			{
				if (Symbol_table.retrieve( paramType->getText() ) != 0)
				{
					if(Symbol_table.retrieve( paramType->getText() )->Type != declared)
					{
    	    	    	stringstream error;
						error << "Parameter " << paramCount << 
							" is not the same type as the declared function parameter";
		
						ErrorHandler::error( paramType->getLine(), error.str());
					}
				}
			}
			else
			{
				if (paramType->getResultType() != declared)
				{
        	    	stringstream error;
					error << "Parameter " << paramCount << 
						" is not the same type as the declared function parameter";

					ErrorHandler::error( paramType->getLine(), error.str());
				}
			}
		}
	})*
	{
		if (entry != 0 && paramCount < entry->paramSize())
		{
		  	stringstream error;
			error << "Function caller has too few parameters ";
			ErrorHandler::error( paramType->getLine(), error.str());
		}
    }
	)
    ;

/**
 * \brief Checks an assign operation.
 *
 * This function checks:
 * - Left hand side of an assignment must be one identifier. 
 * - Const variables are not modified.
 * - The types of the lh and rh are not conflicting.
 */
assign
	:#(op:ASSIGN lh:expression rh:expression)
	{
		if (lh->getType() != IDENT && lh->getType() != ARRAY )
		{
            ErrorHandler::error( lh->getLine(),
                                 "Left hand side of the assignment must be _one_ identifier" );
		}
        else
        {
            // check for const correctness
            // no need to report undeclared var as IDENT does it for us
            if(  Symbol_table.retrieve( lh->getText() ) != 0 &&
                 Symbol_table.retrieve( lh->getText() )->Const == true )
            {
                ErrorHandler::error( lh->getLine(),
                                 "Invalid use of const identifier: " + lh->getText() );
            }
        }
        // check that both expressions are of the same type
        if( lh->getResultType() != rh->getResultType() )
        {
            ErrorHandler::error( op->getLine(), "No match for "
                                 + lh->getResultType().toString()
                                 + " = " + rh->getResultType().toString()
                                 + " expression" );
        }
        else
        {
            op->setResultType( lh->getResultType() );
        }
	}
	;

/**
 * \brief Primitive types, no checking done at this level.
 */
primitive
	:	"void"
	|	"bool"
	|	"char"
	|	"int"
	;


/**
 * \brief Check the get statement.
 *
 * Check if the given parameter is a declared variable and if this function
 * supports it's type. Only boolean, characters and integers can be read. 
 *
 * \bug No array types are supported at the moment. It would be cool if we
 * can read array's of characters. So we can emulate strings.
 *
 * And ofcourse it's return type is set.
 */
getStatement
	: #(getOp:GET
        {
            unsigned int paramCount = 0;
        }(id:variable
        {
            paramCount++;
            // Check for declaration.
            if( Symbol_table.retrieve( id->getText() ) == 0 )
            {
                ErrorHandler::error( id->getLine(), "Undefined variable " + id->getText() );
            }
            else
            {
                // check that the program doesn't try to read some custom type

                if( id->getResultType() != PrimBool &&
                    id->getResultType() != PrimChar &&
                    id->getResultType() != PrimInt )
                    ErrorHandler::error( getOp->getLine(), "No match for get or read: " +
                                         id->getResultType().toString() );
                // check for const correctness
                if( Symbol_table.retrieve( id->getText() )->Const == true )
                {
                    ErrorHandler::error( id->getLine(),
                                         "Invalid use of const identifier: " + id->getText() +
                                         " in put()" );
                }
            }
        } )+ )
		{
            if( paramCount == 1 )
                getOp->setResultType( id->getResultType() );
            else
                getOp->setResultType( PrimVoid );

		}
	;

/**
 * \brief Check the put statement.
 *
 * Check if the given parameter is a declared variable and if this function
 * supports it's type. Only boolean, characters and integers can be written 
 * to screen. 
 * 
 * And ofcourse it's return type is set.
 */
putStatement
    : #(putOp:PUT
              {
                  unsigned int paramCount = 0;
              } (expr:expression /* for each expression check resulttype */
              {
                  paramCount++;
                  if( expr->getResultType() != PrimBool &&
                      expr->getResultType() != PrimChar &&
                      expr->getResultType() != PrimInt )
                      ErrorHandler::error( expr->getLine(), "No match for put or print: " +
                                           expr->getResultType().toString() );
              })+ )
		{
            if( paramCount == 1 )
                putOp->setResultType( expr->getResultType() );
            else
                putOp->setResultType( PrimVoid );
		}
	;

/**
 * \brief Check the if statement.
 *
 * Set it's return type. Default it's void. If the 'if expression' also contains a else statement
 * and both expression blocks have the same type, then (and only then) the return type is
 * the same as the expression block type.
 * 
 * \todo Vars declared in expression should be valid in the two blocks. This is highly illogical if you ask
 * me. It can easily be solved by either adding a looseBlock statement that doesn't open a scope, or we
 * could simply make a dirtyHack boolean that makes the block skip scope opening.
 * Parser: the two expressions really should not be blocks like they are defined!!
 */
ifStatement
	: #(ifOp:IF
    {
        Symbol_table.openScope();
        if( ifOp->getFirstChild()->getType() == BLOCK )
            (static_cast<RefCkASTNodeBase>(ifOp->getFirstChild()))->setIgnoreScope( true );
    } expr:expression ifDo:block (elseDo:block)?)
    {
        if( expr->getResultType() != PrimBool )
        {
            ErrorHandler::error( ifOp->getLine(), "No match for if( "
                                 + expr->getResultType().toString()
                                 + " ) expression" );
        }
        if( elseDo && ifDo->getResultType() == elseDo->getResultType() )
            ifOp->setResultType( ifDo->getResultType() );
        else
            ifOp->setResultType( PrimVoid );

        try
        {
            Symbol_table.closeScope();
        }catch( SymbolTableException & )
        {
            throw Exception( __FILE__, __LINE__, "Compiler oops: Unmatched scopes" );
        }
    }
	;

/**
 * \brief Check the while statement.
 *
 * The while statement has always return type void.
 */
whileStatement
	: #(whileOp:WHILE
    {
        Symbol_table.openScope();
        if( whileOp->getFirstChild()->getType() == BLOCK )
            (static_cast<RefCkASTNodeBase>(whileOp->getFirstChild()))->setIgnoreScope( true );
    } expr:expression whileDo:block )
    {
        whileOp->setResultType( PrimVoid );
        if( expr->getResultType() != PrimBool )
        {
            ErrorHandler::error( whileOp->getLine(), "No match for while( "
                                 + expr->getResultType().toString()
                                 + " ) expression" );
        }

        try
        {
            Symbol_table.closeScope();
        }catch( SymbolTableException & )
        {
            throw Exception( __FILE__, __LINE__, "Compiler oops: Unmatched scopes" );
        }
    }
	;
